
import logging
from typing import Optional

from brownie import chain
from multicall.utils import await_awaitable
from y import convert
from y.datatypes import AnyAddressType, Block, UsdPrice
from y.networks import Network
from y.utils.logging import yLazyLogger
from y.utils.raw_calls import raw_call_async

logger = logging.getLogger(__name__)

POOLS = {
    Network.BinanceSmartChain: {
        "0x86aFa7ff694Ab8C985b79733745662760e454169": "0xF16D312d119c13dD27fD0dC814b0bCdcaAa62dfD", # Belt.fi bDAI/bUSDC/bUSDT/bBUSD
        "0x9cb73F20164e399958261c289Eb5F9846f4D1404": "0xAEA4f7dcd172997947809CE6F12018a6D5c1E8b6", # 4Belt
    },
}.get(chain.id, {})


@yLazyLogger(logger)
def is_belt_lp(token: AnyAddressType) -> bool:
    address = convert.to_address(token)
    return address in POOLS


@yLazyLogger(logger)
def get_price(token: AnyAddressType, block: Optional[Block] = None) -> UsdPrice:
    return await_awaitable(get_price_async(token, block=block))
    
@yLazyLogger(logger)
async def get_price_async(token: AnyAddressType, block: Optional[Block] = None) -> UsdPrice:
    address = convert.to_address(token)
    pool = POOLS[address]
    virtual_price = await raw_call_async(pool, 'get_virtual_price()', output='int', block=block)
    return UsdPrice(virtual_price / 1e18)



import logging
from typing import List, Optional

from async_lru import alru_cache
from brownie import ZERO_ADDRESS
from multicall import Call
from multicall.utils import await_awaitable, gather
from y.classes.common import ERC20
from y.contracts import has_method_async
from y.datatypes import Address, AnyAddressType, Block, UsdPrice, UsdValue
from y.erc20 import decimals, totalSupplyReadable
from y.exceptions import call_reverted
from y.prices import magic
from y.utils.logging import yLazyLogger
from y.utils.raw_calls import raw_call

logger = logging.getLogger(__name__)


@yLazyLogger(logger)
async def is_pie(token: AnyAddressType) -> bool:
    return await_awaitable(is_pie_async(token))

@yLazyLogger(logger)
@alru_cache(maxsize=None)
async def is_pie_async(token: AnyAddressType) -> bool:
    return await has_method_async(token, "getCap()(uint)")

@yLazyLogger(logger)
def get_price(pie: AnyAddressType, block: Optional[Block] = None) -> UsdPrice:
    return UsdPrice(tvl(pie, block) / totalSupplyReadable(pie, block))

@yLazyLogger(logger)
def get_tokens(pie_address: Address, block: Optional[Block] = None) -> List[ERC20]:
    tokens = Call(pie_address, ['getTokens()(address[])'], [['tokens',None]], block_id=block)()['tokens']
    return [ERC20(t) for t in tokens]

@yLazyLogger(logger)
def get_bpool(pie_address: Address, block: Optional[Block] = None) -> Address:
    try:
        bpool = raw_call(pie_address, 'getBPool()', output='address', block=block)
        return bpool if bpool != ZERO_ADDRESS else pie_address
    except Exception as e:
        if not call_reverted(e):
            raise
        return pie_address

@yLazyLogger(logger)
def tvl(pie_address: Address, block: Optional[Block] = None) -> UsdValue:
    tokens = get_tokens(pie_address, block)
    pool = get_bpool(pie_address, block)
    token_balances = gather([Call(token, ['balanceOf(address)(uint)', pool], block_id=block).coroutine() for token in tokens])
    token_decimals = decimals(tokens, block)
    token_balances = [bal / 10 ** decimal for bal, decimal in zip(token_balances, token_decimals)]
    prices = magic.get_prices(tokens, block, silent=True, dop=1)
    return UsdValue(
        sum(bal * price for bal, price in zip(token_balances, prices))
    )




from typing import List

from brownie import chain
import pytest
from tests.prices.dex.test_uniswap import V1_TOKENS, V2_TOKENS
from tests.prices.lending.test_aave import ATOKENS
from tests.prices.lending.test_compound import CTOKENS
from tests.prices.test_chainlink import FEEDS
from tests.prices.test_popsicle import POPSICLES
from tests.prices.test_synthetix import SYNTHS
from y import convert
from y.contracts import contract_creation_block
from y.exceptions import CantFetchParam, PriceError
from y.prices import magic


START_TESTS_AT_BLOCK = 11_000_000
INCREMENT = 200_000
NUM_ITERATIONS = (chain.height - START_TESTS_AT_BLOCK) // INCREMENT
BLOCKS = [START_TESTS_AT_BLOCK + INCREMENT * i for i in range(NUM_ITERATIONS)]

ALL_TOKENS: List = V1_TOKENS + V2_TOKENS + ATOKENS + CTOKENS + FEEDS + POPSICLES + SYNTHS

# subtraction underflow, not relevant to point of test
ALL_TOKENS.remove('0x892B14321a4FCba80669aE30Bd0cd99a7ECF6aC0')
ALL_TOKENS.remove('0x892B14321a4FCba80669aE30Bd0cd99a7ECF6aC0') 
ALL_TOKENS.remove('0x892b14321a4fcba80669ae30bd0cd99a7ecf6ac0')
ALL_TOKENS.remove('0x892B14321A4FCBA80669AE30BD0CD99A7ECF6AC0')

# ibibBTC, not relevant to point of test
ALL_TOKENS.remove('0x2a867fd776b83e1bd4e13c6611afd2f6af07ea6d') 
ALL_TOKENS.remove('0x2A867FD776B83E1BD4E13C6611AFD2F6AF07EA6D')
ALL_TOKENS.remove('0x2A867fd776B83e1bd4e13C6611AFd2F6af07EA6D')
ALL_TOKENS.remove('0x2A867fd776B83e1bd4e13C6611AFd2F6af07EA6D')

# multicall aggregate failed, not relevant to point of test
ALL_TOKENS.remove('0x4ee15f44c6f0d8d1136c83efd2e8e4ac768954c6')
ALL_TOKENS.remove('0x4EE15F44C6F0D8D1136C83EFD2E8E4AC768954C6')
ALL_TOKENS.remove('0x4EE15f44c6F0d8d1136c83EfD2e8E4AC768954c6')
ALL_TOKENS.remove('0x4EE15f44c6F0d8d1136c83EfD2e8E4AC768954c6')
ALL_TOKENS.remove('0x9baf8a5236d44ac410c0186fe39178d5aad0bb87')
ALL_TOKENS.remove('0x9BAF8A5236D44AC410C0186FE39178D5AAD0BB87')
ALL_TOKENS.remove('0x9baF8a5236d44AC410c0186Fe39178d5AAD0Bb87')
ALL_TOKENS.remove('0x9baF8a5236d44AC410c0186Fe39178d5AAD0Bb87')


@pytest.mark.parametrize('block', BLOCKS)
def test_get_prices(block):
    tokens = relevant_tokens(ALL_TOKENS, block)
    checked_separately = [get_price(token, block) for token in tokens]
    checked_together = magic.get_prices(tokens, block, fail_to_None=True)
    for i in range(len(checked_together)):
        assert checked_together[i] == checked_separately[i], f'magic.get_prices price discrepancy for {tokens[i]}'

def get_price(token, block):
    """ This is just to diagnose issues with the test itself. We may need to exclude certain tokens. """
    try:
        return magic.get_price(token, block, fail_to_None=True)
    except CantFetchParam as e:
        raise CantFetchParam(token, str(e))
    except PriceError:
        return None
    except TypeError as e:
        raise TypeError(token, str(e))
    except ValueError as e:
        raise ValueError(token, str(e))

def relevant_tokens(tokens, block):
    return [
        token for token in tokens
        if contract_creation_block(convert.to_address(token))
        and contract_creation_block(convert.to_address(token)) <= block
    ]




import pytest
from tests.test_constants import STABLECOINS
from tests.fixtures import mutate_address, mutate_contract
from tests.prices.lending.test_aave import ATOKENS
from tests.prices.lending.test_compound import CTOKENS
from tests.prices.test_chainlink import FEEDS
from tests.prices.test_popsicle import POPSICLES
from tests.prices.test_synthetix import SYNTHS
from y.constants import EEE_ADDRESS, WRAPPED_GAS_COIN
from y.prices.utils.buckets import check_bucket


@pytest.mark.parametrize('token',ATOKENS)
def test_check_bucket_aave(token):
    assert check_bucket(token) == 'atoken'

@pytest.mark.parametrize('token',FEEDS)
def test_check_bucket_chainlink(token):
    if token in [stable for stable in STABLECOINS if not isinstance(stable,int)]:
        pytest.skip(f'Not applicable to stablecoins.')
    if token in mutate_contract(WRAPPED_GAS_COIN) + mutate_address(EEE_ADDRESS):
        pytest.skip(f'Not applicable to native token.')
    assert check_bucket(token) == 'chainlink feed'

@pytest.mark.parametrize('token',CTOKENS)
def test_check_bucket_compound(token):
    assert check_bucket(token) == 'compound'

@pytest.mark.parametrize('token',POPSICLES)
def test_check_bucket_popsicle(token):
    assert check_bucket(token) == 'popsicle'

@pytest.mark.parametrize('token',STABLECOINS)
def test_check_bucket_stablecoins(token):
    assert check_bucket(token) == 'stable usd'

@pytest.mark.parametrize('token',SYNTHS)
def test_check_bucket_synthetix(token):
    assert check_bucket(token) == 'synthetix'
